import abc

import numpy as np
from scipy.spatial import distance

class BaseKernel(abc.ABC):

    """ A Gaussian Process kernel.

       Attributes:
           hypers (list)
    """

    @abc.abstractmethod
    def __init__(self):
        """ Create a GPKernel. """
        self._n_hypers = 0

    @abc.abstractmethod
    def cov(self, X1, X2, hypers=None):
        """ Calculate the covariance. """
        return np.zeros((len(X1), len(X2)))

    def num_hypers(self):
        """ Return the number of required hyper-parameters. """
        return self._n_hypers

class PolynomialKernel(BaseKernel):

    """ A Polynomial kernel of the form (s0^2 + sp^2 * x.T*x)^d

    Attributes:
    hypers (list): names of the hyperparameters required
    _deg (integer): degree of polynomial
    """

    def __init__(self, d):
        """ Initiate a polynomial kernel.

        Parameters:
            d (integer): degree of the polynomial
        """
        if not isinstance(d, int):
            raise TypeError('d must be an integer.')
        if d < 1:
            raise ValueError('d must be greater than or equal to 1.')
        self._n_hypers = 2
        self._deg = d

    def cov(self, X1, X2, hypers=(1.0, 1.0)):
        """ Calculate the polynomial covariance matrix between X1 and X2.

        Parameters:
            X1 (np.ndarray):
            X2 (np.ndarray):
            hypers (iterable):

        Returns:
            K (np.ndarray)
        """
        ##################
        # Problem 1c
        pass
        ##################


class MaternKernel(BaseKernel):

    """ A Matern kernel with nu = 5/2. """

    def __init__(self):
        """ Initiate a Matern kernel.  """
        self._n_hypers = 1


    def cov(self, X1, X2, hypers=(1.0, )):
        """ Calculate the Matern kernel between X1 and X2.

        Parameters:
            X1 (np.ndarray):
            X2 (np.ndarray)
            hypers (iterable): default is ell=1.0.

        Returns:
            K (np.ndarray)
        """
        ##################
        # Problem 1e
        pass
        ##################



class SEKernel(BaseKernel):

    """ A squared exponential kernel.

    Attributes:
        hypers (list)
        _d_squared (np.ndarray)
        _saved_X (dict)
    """

    def __init__(self):
        """ Initiate a SEKernel. """
        self._n_hypers = 2

    def cov(self, X1, X2, hypers=(1.0, 1.0)):
        """ Calculate the squared exponential kernel between x1 and x2.

        Parameters:
            X1 (np.ndarray):
            X2 (np.ndarray)
            hypers (iterable): default is (1.0, 1.0)

        Returns:
            K (np.ndarray)
        """
        ##################
        # Problem 1d
        pass
        ##################
