import numpy as np
from scipy import optimize, linalg

import kernels

class GPRegressor(object):

    def __init__(self, kernel, hypers=None):
        """ Initialize the Gaussian process regression model.

        The model stores its hyperparameters as an array with the noise
        hyperparameter first and any kernel hyperparameters after.

        Parameters:
            kernel: a kernel from kernels.py
            hypers (np.ndarray): initial hyperparameters; should be 1-D
        """
        self.kernel = kernel
        if hypers is not None:
            self.hypers = hypers
            if len(hypers) != self.kernel.num_hypers() + 1:
                raise ValueError('Incorrect number of initial hyperparameters.')
        else:
            self.hypers = [1.0 for _ in range(self.kernel.num_hypers() + 1)]
            self.hypers = np.array(self.hypers)

    def fit(self, X, y):
        """ Find optimal hyperparameters.

        This method should find the hyperparameters that minimize
        the negative log marginal likelihood using self.hypers as an
        initial guess. It then sets self.hypers to the optimal hyperparameters.

        Parameters:
            X (np.ndarray): training inputs
            y (np.ndarray): training observations
        """
        ##################
        # Problem 3c
        pass
        ##################


    def predict(self, X, y, X_p):
        """ Gaussian process regression predictions.

        Parameters:
            X (np.ndarray): n x d training inputs
            y (np.ndarray): n training observations
            X_p (np.ndarray): m x d points to predict

        Returns:
            mu (np.ndarray): m predicted means
            var (np.ndarray): m predictive variances

        Follows Algorithm 2.1 from GPML.
        """
        ##################
        # Problem 2a
        pass
        ##################

    def marginal_nll_loss(self, log_hypers, X, y):
        """ Calculate the negative log marginal likelihood loss.

        We pass the log hypers here because it makes the optimization
        more stable.

        Parameters:
            log_hypers (np.ndarray): natural log of the hyper-parameters
            X (np.ndarray)
            y (np.ndarray)

        Returns:
            (float) The negative log marginal likelihood.
        """
        ##################
        # Problem 3b
        pass
        ##################
